﻿using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Input;
using System.Windows.Markup;
using System.Windows.Media;
using System.Windows.Threading;

namespace TestLibrary
{
    /// <summary>
    /// MDI 자식
    /// </summary>
    [ContentProperty("Content")]
    public class MDIChild : Control
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// ROUTED EVENT
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 종료 전 이벤트 - ClosingEvent

        /// <summary>
        /// 종료 전 이벤트
        /// </summary>
        public static readonly RoutedEvent ClosingEvent = EventManager.RegisterRoutedEvent
        (
            "Closing",
            RoutingStrategy.Bubble,
            typeof(ClosingEventArgs),
            typeof(MDIChild)
        );

        #endregion
        #region 종료 후 이벤트 - ClosedEvent

        /// <summary>
        /// 종료 후 이벤트
        /// </summary>
        public static readonly RoutedEvent ClosedEvent = EventManager.RegisterRoutedEvent
        (
            "Closed",
            RoutingStrategy.Bubble,
            typeof(RoutedEventArgs),
            typeof(MDIChild)
        );

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Dependency Property
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 위치 속성 - PositionProperty

        /// <summary>
        /// 위치 속성
        /// </summary>
        public static readonly DependencyProperty PositionProperty = DependencyProperty.Register
        (
            "Position",
            typeof(Point),
            typeof(MDIChild),
            new UIPropertyMetadata
            (
                new Point(-1, -1),
                new PropertyChangedCallback(PositionPropertyChangedCallback)
            )
        );

        #endregion
        #region 크기 조정 가능 여부 속성 - ResizableProperty

        /// <summary>
        /// 크기 조정 가능 여부 속성
        /// </summary>
        public static readonly DependencyProperty ResizableProperty = DependencyProperty.Register
        (
            "Resizable",
            typeof(bool),
            typeof(MDIChild),
            new UIPropertyMetadata(true)
        );

        #endregion
        #region 윈도우 상태 속성 - WindowStateProperty

        /// <summary>
        /// 윈도우 상태 속성
        /// </summary>
        public static readonly DependencyProperty WindowStateProperty = DependencyProperty.Register
        (
            "WindowState",
            typeof(WindowState),
            typeof(MDIChild),
            new UIPropertyMetadata
            (
                WindowState.Normal,
                new PropertyChangedCallback(WindowStatePropertyChangedCallback)
            )
        );

        #endregion
        #region 포커스 여부 - FocusedProperty

        /// <summary>
        /// 포커스 여부
        /// </summary>
        public static readonly DependencyProperty FocusedProperty = DependencyProperty.Register
        (
            "Focused",
            typeof(bool),
            typeof(MDIChild),
            new UIPropertyMetadata
            (
                false,
                new PropertyChangedCallback(FocusedPropertyChangedCallback)
            )
        );

        #endregion
        #region 아이콘 표시 여부 속성 - ShowIconProperty

        /// <summary>
        /// 아이콘 표시 여부 속성
        /// </summary>
        public static readonly DependencyProperty ShowIconProperty = DependencyProperty.Register
        (
            "ShowIcon",
            typeof(bool),
            typeof(MDIChild),
            new UIPropertyMetadata(true)
        );

        #endregion
        #region 아이콘 속성 - IconProperty

        /// <summary>
        /// 아이콘 속성
        /// </summary>
        public static readonly DependencyProperty IconProperty = DependencyProperty.Register
        (
            "Icon",
            typeof(ImageSource),
            typeof(MDIChild)
        );

        #endregion
        #region 제목 속성 - TitleProperty

        /// <summary>
        /// 제목 속성
        /// </summary>
        public static readonly DependencyProperty TitleProperty = DependencyProperty.Register
        (
            "Title",
            typeof(string),
            typeof(MDIChild)
        );

        #endregion
        #region 최소화 상자 속성 - MinimizeBoxProperty

        /// <summary>
        /// 최소화 상자 속성
        /// </summary>
        public static readonly DependencyProperty MinimizeBoxProperty = DependencyProperty.Register
        (
            "MinimizeBox",
            typeof(bool),
            typeof(MDIChild),
            new UIPropertyMetadata
            (
                true,
                new PropertyChangedCallback(MinimizeBoxPropertyChangedCallback)
            )
        );

        #endregion
        #region 최대화 상자 속성 - MaximizeBoxProperty

        /// <summary>
        /// 최대화 상자 속성
        /// </summary>
        public static readonly DependencyProperty MaximizeBoxProperty = DependencyProperty.Register
        (
            "MaximizeBox",
            typeof(bool),
            typeof(MDIChild),
            new UIPropertyMetadata
            (
                true,
                new PropertyChangedCallback(MaximizeBoxPropertyChangedCallback)
            )
        );

        #endregion
        #region 닫기 상자 속성 - CloseBoxProperty

        /// <summary>
        /// 닫기 상자 속성
        /// </summary>
        public static readonly DependencyProperty CloseBoxProperty = DependencyProperty.Register
        (
            "CloseBox",
            typeof(bool),
            typeof(MDIChild),
            new UIPropertyMetadata
            (
                true,
                new PropertyChangedCallback(CloseBoxPropertyChangedCallback)
            )
        );

        #endregion
        #region 컨텐트 속성 - ContentProperty

        /// <summary>
        /// 컨텐트 속성
        /// </summary>
        public static readonly DependencyProperty ContentProperty = DependencyProperty.Register
        (
            "Content",
            typeof(UIElement),
            typeof(MDIChild)
        );

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Event
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 종료 전 이벤트 - Closing

        /// <summary>
        /// 종료 전 이벤트
        /// </summary>
        public event RoutedEventHandler Closing
        {
            add
            {
                AddHandler(ClosingEvent, value);
            }
            remove
            {
                RemoveHandler(ClosingEvent, value);
            }
        }

        #endregion
        #region 종료 후 이벤트 - Closed

        /// <summary>
        /// 종료 후 이벤트
        /// </summary>
        public event RoutedEventHandler Closed
        {
            add
            {
                AddHandler(ClosedEvent, value);
            }
            remove
            {
                RemoveHandler(ClosedEvent, value);
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Internal

        #region Field

        /// <summary>
        /// 최소화시 너비
        /// </summary>
        internal const int MinimizedWidth = 160;

        /// <summary>
        /// 최소화시 높이
        /// </summary>
        internal const int MinimizedHeight = 29;

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 최소화 버튼
        /// </summary>
        private Button minimizeButton;

        /// <summary>
        /// 최대화 버튼
        /// </summary>
        private Button maximizeButton;

        /// <summary>
        /// 닫기 버튼
        /// </summary>
        private Button closeButton;

        /// <summary>
        /// 버튼 스택 패널
        /// </summary>
        private StackPanel buttonStackPanel;

        /// <summary>
        /// 원본 치수
        /// </summary>
        private Rect originalDimension;

        /// <summary>
        /// 최소화 위치
        /// </summary>
        private Point minimizedPosition = new Point(-1, -1);

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 위치 - Position

        /// <summary>
        /// 위치
        /// </summary>
        public Point Position
        {
            get
            {
                return (Point)GetValue(PositionProperty);
            }
            set
            {
                SetValue(PositionProperty, value);
            }
        }

        #endregion
        #region 크기 조정 가능 여부 - Resizable

        /// <summary>
        /// 크기 조정 가능 여부
        /// </summary>
        public bool Resizable
        {
            get
            {
                return (bool)GetValue(ResizableProperty);
            }
            set
            {
                SetValue(ResizableProperty, value);
            }
        }

        #endregion
        #region 윈도우 상태 - WindowState

        /// <summary>
        /// 윈도우 상태
        /// </summary>
        public WindowState WindowState
        {
            get
            {
                return (WindowState)GetValue(WindowStateProperty);
            }
            set
            {
                SetValue(WindowStateProperty, value);
            }
        }

        #endregion
        #region 포커스 여부 - Focused

        /// <summary>
        /// 포커스 여부
        /// </summary>
        public bool Focused
        {
            get
            {
                return (bool)GetValue(FocusedProperty);
            }
            set
            {
                SetValue(FocusedProperty, value);
            }
        }

        #endregion
        #region 아이콘 표시 여부 - ShowIcon

        /// <summary>
        /// 아이콘 표시 여부
        /// </summary>
        public bool ShowIcon
        {
            get
            {
                return (bool)GetValue(ShowIconProperty);
            }
            set
            {
                SetValue(ShowIconProperty, value);
            }
        }

        #endregion
        #region 아이콘 - Icon

        /// <summary>
        /// 아이콘
        /// </summary>
        public ImageSource Icon
        {
            get
            {
                return (ImageSource)GetValue(IconProperty);
            }
            set
            {
                SetValue(IconProperty, value);
            }
        }

        #endregion
        #region 제목 - Title

        /// <summary>
        /// 제목
        /// </summary>
        public string Title
        {
            get
            {
                return (string)GetValue(TitleProperty);
            }
            set
            {
                SetValue(TitleProperty, value);
            }
        }

        #endregion
        #region 최소화 상자 - MinimizeBox

        /// <summary>
        /// 최소화 상자
        /// </summary>
        public bool MinimizeBox
        {
            get
            {
                return (bool)GetValue(MinimizeBoxProperty);
            }
            set
            {
                SetValue(MinimizeBoxProperty, value);
            }
        }

        #endregion
        #region 최대화 상자 - MaximizeBox

        /// <summary>
        /// 최대화 상자
        /// </summary>
        public bool MaximizeBox
        {
            get
            {
                return (bool)GetValue(MaximizeBoxProperty);
            }
            set
            {
                SetValue(MaximizeBoxProperty, value);
            }
        }

        #endregion
        #region 닫기 상자 - CloseBox

        /// <summary>
        /// 닫기 상자
        /// </summary>
        public bool CloseBox
        {
            get
            {
                return (bool)GetValue(CloseBoxProperty);
            }
            set
            {
                SetValue(CloseBoxProperty, value);
            }
        }

        #endregion
        #region 컨텐트 - Content

        /// <summary>
        /// 컨텐트
        /// </summary>
        public UIElement Content
        {
            get
            {
                return (UIElement)GetValue(ContentProperty);
            }
            set
            {
                SetValue(ContentProperty, value);
            }
        }

        #endregion

        #region 컨테이너 - Container

        /// <summary>
        /// 컨테이너
        /// </summary>
        public MDIContainer Container { get; private set; }

        #endregion
        #region 버튼 패널 - ButtonPanel

        /// <summary>
        /// 버튼 패널
        /// </summary>
        public Panel ButtonPanel { get; private set; }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 비 최대화 상태 - NonMaximizedState

        /// <summary>
        /// 비 최대화 상태
        /// </summary>
        private WindowState NonMaximizedState { get; set; }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Static

        #region 생성자 - MDIChild()

        /// <summary>
        /// 생성자
        /// </summary>
        static MDIChild()
        {
            DefaultStyleKeyProperty.OverrideMetadata
            (
                typeof(MDIChild),
                new FrameworkPropertyMetadata(typeof(MDIChild))
            );
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - MDIChild()

        /// <summary>
        /// 생성자
        /// </summary>
        public MDIChild()
        {
            Focusable = IsTabStop = false;

            Loaded   += Control_Loaded;
            GotFocus += Control_GotFocus;
            KeyDown  += Control_KeyDown;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static

        #region 위치 속성 변경시 콜백 처리하기 - PositionPropertyChangedCallback(d, e)

        /// <summary>
        /// 위치 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void PositionPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            if((Point)e.NewValue == (Point)e.OldValue)
            {
                return;
            }

            MDIChild mdiChild = d as MDIChild;

            Point newPosition = (Point)e.NewValue;

            Canvas.SetTop (mdiChild, newPosition.Y < 0 ? 0 : newPosition.Y);
            Canvas.SetLeft(mdiChild, newPosition.X < 0 ? 0 : newPosition.X);
        }

        #endregion
        #region 윈도우 상태 속성 변경시 콜백 처리하기 - WindowStatePropertyChangedCallback(d, e)

        /// <summary>
        /// 윈도우 상태 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void WindowStatePropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            MDIChild     mdiChild     = d as MDIChild;
            MDIContainer mdiContainer = mdiChild.Container;

            WindowState previousWindowState = (WindowState)e.OldValue;
            WindowState windowState         = (WindowState)e.NewValue;

            if(mdiChild.Container == null || previousWindowState == windowState)
            {
                return;
            }

            if(previousWindowState == WindowState.Maximized)
            {
                if(mdiContainer.ActiveMDIChild.WindowState != WindowState.Maximized)
                {
                    for(int i = 0; i < mdiContainer.Children.Count; i++)
                    {
                        if
                        (
                            mdiContainer.Children[i] != mdiChild &&
                            mdiContainer.Children[i].WindowState == WindowState.Maximized &&
                            mdiContainer.Children[i].MaximizeBox
                        )
                        {
                            mdiContainer.Children[i].WindowState = WindowState.Normal;
                        }
                    }

                    ScrollViewer scrollViewer = (ScrollViewer)((Grid)mdiContainer.Content).Children[1];

                    scrollViewer.HorizontalScrollBarVisibility = ScrollBarVisibility.Auto;
                    scrollViewer.VerticalScrollBarVisibility   = ScrollBarVisibility.Auto;
                }

                mdiChild.ButtonPanel.Children.Clear();

                mdiChild.ButtonPanel = null;

                mdiChild.buttonStackPanel.Children.Add(mdiChild.minimizeButton);
                mdiChild.buttonStackPanel.Children.Add(mdiChild.maximizeButton);
                mdiChild.buttonStackPanel.Children.Add(mdiChild.closeButton   );
            }

            if(previousWindowState == WindowState.Minimized)
            {
                mdiChild.minimizedPosition = mdiChild.Position;
            }

            switch(windowState)
            {
                case WindowState.Normal :
                {
                    mdiChild.Position = new Point(mdiChild.originalDimension.X, mdiChild.originalDimension.Y);
                    mdiChild.Width    = mdiChild.originalDimension.Width;
                    mdiChild.Height   = mdiChild.originalDimension.Height;

                    break;
                }
                case WindowState.Minimized :
                {
                    if(previousWindowState == WindowState.Normal)
                    {
                        mdiChild.originalDimension = new Rect
                        (
                            mdiChild.Position.X,
                            mdiChild.Position.Y,
                            mdiChild.ActualWidth,
                            mdiChild.ActualHeight
                        );
                    }

                    double newLeft;
                    double newTop;

                    if(mdiChild.minimizedPosition.X >= 0 || mdiChild.minimizedPosition.Y >= 0)
                    {
                        newLeft = mdiChild.minimizedPosition.X;
                        newTop  = mdiChild.minimizedPosition.Y;
                    }
                    else
                    {
                        List<Rect> minimizedWindowList = new List<Rect>();

                        for(int i = 0; i < mdiContainer.Children.Count; i++)
                        {
                            MDIChild child = mdiContainer.Children[i];

                            if(child != mdiChild && child.WindowState == WindowState.Minimized)
                            {
                                minimizedWindowList.Add
                                (
                                    new Rect
                                    (
                                        child.Position.X,
                                        mdiContainer.InnerHeight - child.Position.Y,
                                        child.Width,
                                        child.Height
                                    )
                                );
                            }
                        }

                        Rect newWindowRectangle;
                        bool occupied = true;
                        int  count    = 0;
                        int  capacity = Convert.ToInt32(mdiContainer.ActualWidth) / MDIChild.MinimizedWidth;

                        do
                        {
                            int row    = count / capacity + 1;
                            int column = count % capacity;

                            newTop  = MDIChild.MinimizedHeight * row;
                            newLeft = MDIChild.MinimizedWidth * column;

                            newWindowRectangle = new Rect
                            (
                                newLeft,
                                newTop,
                                MDIChild.MinimizedWidth,
                                MDIChild.MinimizedHeight
                            );

                            occupied = false;

                            foreach(Rect rectangle in minimizedWindowList)
                            {
                                Rect intersection = rectangle;

                                intersection.Intersect(newWindowRectangle);

                                if
                                (
                                    intersection != Rect.Empty &&
                                    intersection.Width  > 0    &&
                                    intersection.Height > 0
                                )
                                {
                                    occupied = true;

                                    break;
                                }
                            }

                            count++;

                            if(newTop < 0)
                            {
                                newTop = 0;

                                occupied = false;
                            }
                        }
                        while(occupied);

                        newTop = mdiContainer.InnerHeight - newTop;
                    }

                    mdiChild.Position = new Point(newLeft, newTop);

                    mdiChild.Width  = MDIChild.MinimizedWidth;
                    mdiChild.Height = MDIChild.MinimizedHeight;

                    break;
                }
                case WindowState.Maximized :
                {
                    if(previousWindowState == WindowState.Normal)
                    {
                        mdiChild.originalDimension = new Rect
                        (
                            mdiChild.Position.X,
                            mdiChild.Position.Y,
                            mdiChild.ActualWidth,
                            mdiChild.ActualHeight
                        );
                    }

                    mdiChild.NonMaximizedState = previousWindowState;

                    mdiChild.buttonStackPanel.Children.Clear();

                    StackPanel stackPanel = new StackPanel
                    {
                        Orientation = Orientation.Horizontal
                    };

                    stackPanel.Children.Add(mdiChild.minimizeButton);
                    stackPanel.Children.Add(mdiChild.maximizeButton);
                    stackPanel.Children.Add(mdiChild.closeButton   );

                    mdiChild.Position    = new Point(0, 0);
                    mdiChild.Width       = mdiContainer.ActualWidth;
                    mdiChild.Height      = mdiContainer.InnerHeight - 2;
                    mdiChild.ButtonPanel = stackPanel;

                    ScrollViewer scrollViewer = (ScrollViewer)((Grid)mdiContainer.Content).Children[1];

                    scrollViewer.HorizontalScrollBarVisibility = ScrollBarVisibility.Hidden;
                    scrollViewer.VerticalScrollBarVisibility   = ScrollBarVisibility.Hidden;

                    break;
                }
            }
    
            if(mdiContainer.ActiveMDIChild == mdiChild)
            {
                mdiContainer.ButtonPanel = mdiChild.ButtonPanel;
            }

            mdiContainer.InvalidateSize();
        }

        #endregion
        #region 포커스 여부 속성 변경시 콜백 처리하기 - FocusedPropertyChangedCallback(d, e)

        /// <summary>
        /// 포커스 여부 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void FocusedPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            if((bool)e.NewValue == (bool)e.OldValue)
            {
                return;
            }

            MDIChild mdiChild = (MDIChild)d;

            if((bool)e.NewValue)
            {
                mdiChild.Dispatcher.BeginInvoke
                (
                    new Func<IInputElement, IInputElement>(Keyboard.Focus),
                    DispatcherPriority.ApplicationIdle,
                    mdiChild.Content
                );

                mdiChild.RaiseEvent(new RoutedEventArgs(GotFocusEvent, mdiChild));
            }
            else
            {
                if(mdiChild.WindowState == WindowState.Maximized)
                {
                    mdiChild.Unmaximize();
                }

                mdiChild.RaiseEvent(new RoutedEventArgs(LostFocusEvent, mdiChild));
            }
        }

        #endregion
        #region 최소화 버튼 속성 변경시 콜백 처리하기 - MinimizeBoxPropertyChangedCallback(d, e)

        /// <summary>
        /// 최소화 버튼 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private static void MinimizeBoxPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            MDIChild mdiChild = d as MDIChild;

            bool visible = (bool)e.NewValue;

            if(visible)
            {
                bool maximizeVisible = true;

                if(mdiChild.maximizeButton != null)
                {
                    maximizeVisible = mdiChild.maximizeButton.Visibility == Visibility.Visible;
                }

                if(mdiChild.minimizeButton != null)
                {
                    mdiChild.minimizeButton.IsEnabled = true;
                }

                if(!maximizeVisible)
                {
                    if(mdiChild.maximizeButton != null)
                    {
                        mdiChild.minimizeButton.Visibility = Visibility.Visible;
                    }

                    if(mdiChild.maximizeButton != null)
                    {
                        mdiChild.maximizeButton.Visibility = Visibility.Visible;
                    }
                }
            }
            else
            {
                bool maximizeEnabled = true;

                if(mdiChild.maximizeButton != null)
                {
                    maximizeEnabled = mdiChild.maximizeButton.IsEnabled;
                }

                if(mdiChild.minimizeButton != null)
                {
                    mdiChild.minimizeButton.IsEnabled = false;
                }

                if(!maximizeEnabled)
                {
                    if(mdiChild.minimizeButton != null)
                    {
                        mdiChild.minimizeButton.Visibility = Visibility.Hidden;
                    }

                    if(mdiChild.maximizeButton != null)
                    {
                        mdiChild.maximizeButton.Visibility = Visibility.Hidden;
                    }
                }
            }
        }

        #endregion
        #region 최대화 버튼 속성 변경시 콜백 처리하기 - MaximizeBoxPropertyChangedCallback(d, e)

        /// <summary>
        /// 최대화 버튼 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private static void MaximizeBoxPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            MDIChild mdiChild = d as MDIChild;

            bool visible = (bool)e.NewValue;

            if(visible)
            {
                bool minimizeVisible = true;

                if(mdiChild.minimizeButton != null)
                {
                    minimizeVisible = mdiChild.minimizeButton.Visibility == Visibility.Visible;
                }

                if(mdiChild.maximizeButton != null)
                {
                    mdiChild.maximizeButton.IsEnabled = true;
                }

                if(!minimizeVisible)
                {
                    if(mdiChild.maximizeButton != null)
                    {
                        mdiChild.minimizeButton.Visibility = Visibility.Visible;
                    }

                    if(mdiChild.maximizeButton != null)
                    {
                        mdiChild.maximizeButton.Visibility = Visibility.Visible;
                    }
                }
            }
            else
            {
                bool minimizeEnabled = true;

                if(mdiChild.minimizeButton != null)
                {
                    minimizeEnabled = mdiChild.minimizeButton.IsEnabled;
                }

                if(mdiChild.maximizeButton != null)
                {
                    mdiChild.maximizeButton.IsEnabled = false;
                }

                if(!minimizeEnabled)
                {
                    if(mdiChild.maximizeButton != null)
                    {
                        mdiChild.minimizeButton.Visibility = Visibility.Hidden;
                    }

                    if(mdiChild.maximizeButton != null)
                    {
                        mdiChild.maximizeButton.Visibility = Visibility.Hidden;
                    }
                }
            }
        }

        #endregion
        #region 닫기 버튼 속성 변경시 콜백 처리하기 - CloseBoxPropertyChangedCallback(d, e)

        /// <summary>
        /// 닫기 버튼 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private static void CloseBoxPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            MDIChild mdiChild = d as MDIChild;

            bool visible = (bool)e.NewValue;

            if(visible)
            {
                if((mdiChild.closeButton != null) && (mdiChild.closeButton.Visibility != Visibility.Visible))
                {
                    mdiChild.closeButton.Visibility = Visibility.Visible;
                }
            }
            else
            {
                if((mdiChild.closeButton != null) && (mdiChild.closeButton.Visibility == Visibility.Visible))
                {
                    mdiChild.closeButton.Visibility = Visibility.Collapsed;
                }
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Internal
        ////////////////////////////////////////////////////////////////////// Function

        #region 비최대화 처리하기 - Unmaximize()

        /// <summary>
        /// 비최대화 처리하기
        /// </summary>
        internal void Unmaximize()
        {
            if(WindowState == WindowState.Maximized)
            {
                WindowState = NonMaximizedState;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Public
        ////////////////////////////////////////////////////////////////////// Function

        #region 포커스 설정하기 - Focus()

        /// <summary>
        /// 포커스 설정하기
        /// </summary>
        public new void Focus()
        {
            Container.ActiveMDIChild = this;
        }

        #endregion
        #region 닫기 - Close()

        /// <summary>
        /// 닫기
        /// </summary>
        public void Close()
        {
            if(ButtonPanel != null)
            {
                ButtonPanel.Children.Clear();
            }

            Container.Children.Remove(this);
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Protected
        ////////////////////////////////////////////////////////////////////// Function

        #region 템플리트 적용시 처리하기 - OnApplyTemplate()

        /// <summary>
        /// 템플리트 적용시 처리하기
        /// </summary>
        public override void OnApplyTemplate()
        {
            base.OnApplyTemplate();

            this.minimizeButton = Template.FindName("MinimizeButton", this) as Button;
            this.maximizeButton = Template.FindName("MaximizeButton", this) as Button;
            this.closeButton    = Template.FindName("CloseButton"   , this) as Button;

            this.buttonStackPanel = Template.FindName("ButtonStackPanel", this) as StackPanel;

            if(this.minimizeButton != null)
            {
                this.minimizeButton.Click += minimizeButton_Click;
            }

            if(this.maximizeButton != null)
            {
                this.maximizeButton.Click += maximizeButton_Click;
            }

            if(this.closeButton != null)
            {
                this.closeButton.Click += closeButton_Click;
            }

            Thumb dragThumb = Template.FindName("DragThumb", this) as Thumb;

            if(dragThumb != null)
            {
                dragThumb.DragStarted += Thumb_DragStarted;
                dragThumb.DragDelta   += dragThumb_DragDelta;

                dragThumb.MouseDoubleClick += (sender, e) =>
                {
                    if(WindowState == WindowState.Minimized)
                    {
                        minimizeButton_Click(null, null);
                    }
                    else if(WindowState == WindowState.Normal)
                    {
                        maximizeButton_Click(null, null);
                    }
                };
            }

            Thumb resizeLeftThumb        = Template.FindName("ResizeLeft"       , this) as Thumb;
            Thumb resizeTopLeftThumb     = Template.FindName("ResizeTopLeft"    , this) as Thumb;
            Thumb resizeTopThumb         = Template.FindName("ResizeTop"        , this) as Thumb;
            Thumb resizeTopRightThumb    = Template.FindName("ResizeTopRight"   , this) as Thumb;
            Thumb resizeRightThumb       = Template.FindName("ResizeRight"      , this) as Thumb;
            Thumb resizeBottomRightThumb = Template.FindName("ResizeBottomRight", this) as Thumb;
            Thumb resizeBottomThumb      = Template.FindName("ResizeBottom"     , this) as Thumb;
            Thumb resizeBottomLeftThumb  = Template.FindName("ResizeBottomLeft" , this) as Thumb;

            if(resizeLeftThumb != null)
            {
                resizeLeftThumb.DragStarted += Thumb_DragStarted;
                resizeLeftThumb.DragDelta   += resizeLeftThumb_DragDelta;
            }

            if(resizeTopThumb != null)
            {
                resizeTopThumb.DragStarted += Thumb_DragStarted;
                resizeTopThumb.DragDelta   += resizeTopThumb_DragDelta;
            }

            if(resizeRightThumb != null)
            {
                resizeRightThumb.DragStarted += Thumb_DragStarted;
                resizeRightThumb.DragDelta   += resizeRightThumb_DragDelta;
            }

            if(resizeBottomThumb != null)
            {
                resizeBottomThumb.DragStarted += Thumb_DragStarted;
                resizeBottomThumb.DragDelta   += resizeBottomThumb_DragDelta;
            }

            if(resizeTopLeftThumb != null)
            {
                resizeTopLeftThumb.DragStarted += Thumb_DragStarted;

                resizeTopLeftThumb.DragDelta += (sender, e) =>
                {
                    resizeTopThumb_DragDelta(null, e);
                    resizeLeftThumb_DragDelta(null, e);

                    Container.InvalidateSize();
                };
            }

            if(resizeTopRightThumb != null)
            {
                resizeTopRightThumb.DragStarted += Thumb_DragStarted;

                resizeTopRightThumb.DragDelta += (sender, e) =>
                {
                    resizeTopThumb_DragDelta(null, e);
                    resizeRightThumb_DragDelta(null, e);

                    Container.InvalidateSize();
                };
            }

            if(resizeBottomRightThumb != null)
            {
                resizeBottomRightThumb.DragStarted += Thumb_DragStarted;

                resizeBottomRightThumb.DragDelta += (sender, e) =>
                {
                    resizeBottomThumb_DragDelta(null, e);
                    resizeRightThumb_DragDelta(null, e);

                    Container.InvalidateSize();
                };
            }

            if(resizeBottomLeftThumb != null)
            {
                resizeBottomLeftThumb.DragStarted += Thumb_DragStarted;

                resizeBottomLeftThumb.DragDelta += (sender, e) =>
                {
                    resizeBottomThumb_DragDelta(null, e);
                    resizeLeftThumb_DragDelta(null, e);

                    Container.InvalidateSize();
                };
            }

            MinimizeBoxPropertyChangedCallback
            (
                this,
                new DependencyPropertyChangedEventArgs(MinimizeBoxProperty, true, MinimizeBox)
            );

            MaximizeBoxPropertyChangedCallback
            (
                this,
                new DependencyPropertyChangedEventArgs(MaximizeBoxProperty, true, MaximizeBox)
            );

            CloseBoxPropertyChangedCallback
            (
                this,
                new DependencyPropertyChangedEventArgs(CloseBoxProperty, true, CloseBox)
            );
        }

        #endregion
        #region 마우스 DOWN 처리하기 - OnMouseDown(e)

        /// <summary>
        /// 마우스 DOWN 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnMouseDown(MouseButtonEventArgs e)
        {
            base.OnMouseDown(e);

            Focused = true;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Private
        ////////////////////////////////////////////////////////////////////// Event

        #region 컨트롤 로드시 처리하기 - Control_Loaded(sender, e)

        /// <summary>
        /// 컨트롤 로드시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void Control_Loaded(object sender, RoutedEventArgs e)
        {
            FrameworkElement frameworkElement = this;

            while(frameworkElement != null && frameworkElement.GetType() != typeof(MDIContainer))
            {
                frameworkElement = frameworkElement.Parent as FrameworkElement;
            }

            if(frameworkElement != null)
            {
                Container = frameworkElement as MDIContainer;
            }
        }

        #endregion
        #region 컨트롤 키 DOWN 처리하기 - Control_KeyDown(sender, e)

        /// <summary>
        /// 컨트롤 키 DOWN 처리하기
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Control_KeyDown(object sender, KeyEventArgs e)
        {
            MDIChild child = sender as MDIChild;

            switch(e.Key)
            {
                case Key.F4 :

                    if(e.KeyboardDevice.Modifiers == ModifierKeys.Control)
                    {
                        child.Close();

                        e.Handled = true;
                    }

                    break;
            }
        }

        #endregion
        #region 컨트롤 포커스 획득시 처리하기 - Control_GotFocus(sender, e)

        /// <summary>
        /// 컨트롤 포커스 획득시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void Control_GotFocus(object sender, RoutedEventArgs e)
        {
            Focus();
        }

        #endregion

        #region 최소화 버튼 클릭시 처리하기 - minimizeButton_Click(sender, e)

        /// <summary>
        /// 최소화 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void minimizeButton_Click(object sender, RoutedEventArgs e)
        {
            if(WindowState == WindowState.Minimized)
            {
                WindowState = WindowState.Normal;
            }
            else
            {
                WindowState = WindowState.Minimized;
            }
        }

        #endregion
        #region 최대화 버튼 클릭시 처리하기 - maximizeButton_Click(sender, e)

        /// <summary>
        /// 최대화 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void maximizeButton_Click(object sender, RoutedEventArgs e)
        {
            if(WindowState == WindowState.Maximized)
            {
                WindowState = WindowState.Normal;
            }
            else
            {
                WindowState = WindowState.Maximized;
            }
        }

        #endregion
        #region 닫기 버튼 클릭시 처리하기 - closeButton_Click(sender, e)

        /// <summary>
        /// 닫기 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void closeButton_Click(object sender, RoutedEventArgs e)
        {
            ClosingEventArgs eventArgs = new ClosingEventArgs(ClosingEvent);

            RaiseEvent(eventArgs);

            if(eventArgs.Cancel)
            {
                return;
            }

            Close();

            RaiseEvent(new RoutedEventArgs(ClosedEvent));
        }

        #endregion

        #region THUMB 드래그 시작시 처리하기 - Thumb_DragStarted(sender, e)

        /// <summary>
        /// THUMB 드래그 시작시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void Thumb_DragStarted(object sender, DragStartedEventArgs e)
        {
            if(!Focused)
            {
                Focused = true;
            }
        }

        #endregion
        #region 드래그 THUMB 드래그 변화시 처리하기 - dragThumb_DragDelta(sender, e)

        /// <summary>
        /// 드래그 THUMB 드래그 변화시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void dragThumb_DragDelta(object sender, DragDeltaEventArgs e)
        {
            if(WindowState == WindowState.Maximized)
            {
                return;
            }

            double newLeft = Position.X + e.HorizontalChange;
            double newTop  = Position.Y + e.VerticalChange;

            if(newLeft < 0)
            {
                newLeft = 0;
            }

            if(newTop < 0)
            {
                newTop = 0;
            }

            Position = new Point(newLeft, newTop);

            Container.InvalidateSize();
        }

        #endregion
        #region 크기 변경 왼쪽 THUMB 드래그 변화시 처리하기 - resizeLeftThumb_DragDelta(sender, e)

        /// <summary>
        /// 크기 변경 왼쪽 THUMB 드래그 변화시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void resizeLeftThumb_DragDelta(object sender, DragDeltaEventArgs e)
        {
            if(Width - e.HorizontalChange < MinWidth)
            {
                return;
            }

            double newLeft = e.HorizontalChange;

            if(Position.X + newLeft < 0)
            {
                newLeft = 0 - Position.X;
            }

            Width -= newLeft;

            Position = new Point(Position.X + newLeft, Position.Y);

            if(sender != null)
            {
                Container.InvalidateSize();
            }
        }

        #endregion
        #region 크기 변경 위쪽 THUMB 드래그 변화시 처리하기 - resizeTopThumb_DragDelta(sender, e)

        /// <summary>
        /// 크기 변경 위쪽 THUMB 드래그 변화시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void resizeTopThumb_DragDelta(object sender, DragDeltaEventArgs e)
        {
            if(Height - e.VerticalChange < MinHeight)
            {
                return;
            }

            double newTop = e.VerticalChange;

            if(Position.Y + newTop < 0)
            {
                newTop = 0 - Position.Y;
            }

            Height -= newTop;

            Position = new Point(Position.X, Position.Y + newTop);

            if(sender != null)
            {
                Container.InvalidateSize();
            }
        }

        #endregion
        #region 크기 변경 오른쪽 THUMB 드래그 변화시 처리하기 - resizeRightThumb_DragDelta(sender, e)

        /// <summary>
        /// 크기 변경 오른쪽 THUMB 드래그 변화시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void resizeRightThumb_DragDelta(object sender, DragDeltaEventArgs e)
        {
            if(Width + e.HorizontalChange < MinWidth)
            {
                return;
            }

            Width += e.HorizontalChange;

            if(sender != null)
            {
                Container.InvalidateSize();
            }
        }

        #endregion
        #region 크기 변경 위쪽 THUMB 드래그 변화시 처리하기 - resizeBottomThumb_DragDelta(sender, e)

        /// <summary>
        /// 크기 변경 위쪽 THUMB 드래그 변화시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void resizeBottomThumb_DragDelta(object sender, DragDeltaEventArgs e)
        {
            if(Height + e.VerticalChange < MinHeight)
            {
                return;
            }

            Height += e.VerticalChange;

            if(sender != null)
            {
                Container.InvalidateSize();
            }
        }

        #endregion
    }
}