﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Markup;
using System.Windows.Media;

namespace TestLibrary
{
    /// <summary>
    /// MDI 컨테이너
    /// </summary>
    [ContentProperty("Children")]
    public class MDIContainer : UserControl
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Class
        ////////////////////////////////////////////////////////////////////////////////////////// Internal

        #region MDI 자식 비교자 - MDIChildComparer

        /// <summary>
        /// MDI 자식 비교자
        /// </summary>
        internal class MDIChildComparer : IComparer<MDIChild>
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////// Method
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 비교하기 - Compare(child1, child2)

            /// <summary>
            /// 비교하기
            /// </summary>
            /// <param name="child1">자식 1</param>
            /// <param name="child2">자식 2</param>
            /// <returns>비교 결과</returns>
            public int Compare(MDIChild child1, MDIChild child2)
            {
                return -1 * Canvas.GetZIndex(child1).CompareTo(Canvas.GetZIndex(child2));
            }

            #endregion
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Event
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region MDI 자식 제목 변경시 이벤트 - MDIChildTitleChanged

        /// <summary>
        /// MDI 자식 제목 변경시 이벤트
        /// </summary>
        public event RoutedEventHandler MDIChildTitleChanged;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Dependency Property
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 테마 속성 - ThemeProperty

        /// <summary>
        /// 테마 속성
        /// </summary>
        public static readonly DependencyProperty ThemeProperty = DependencyProperty.Register
        (
            "Theme",
            typeof(ThemeType),
            typeof(MDIContainer),
            new UIPropertyMetadata
            (
                ThemeType.Aero,
                new PropertyChangedCallback(ThemePropertyChangedCallback)
            )
        );

        #endregion
        #region 메뉴 속성 - MenuProperty

        /// <summary>
        /// 메뉴 속성
        /// </summary>
        public static readonly DependencyProperty MenuProperty = DependencyProperty.Register
        (
            "Menu",
            typeof(UIElement),
            typeof(MDIContainer),
            new UIPropertyMetadata
            (
                null,
                new PropertyChangedCallback(MenuPropertyChangedCallback)
            )
        );

        #endregion
        #region MDI 레이아웃 - MDILayoutProperty

        /// <summary>
        /// MDI 레이아웃
        /// </summary>
        public static readonly DependencyProperty MDILayoutProperty = DependencyProperty.Register
        (
            "MDILayout",
            typeof(MDILayout),
            typeof(MDIContainer),
            new UIPropertyMetadata
            (
                MDILayout.ArrangeIcons,
                new PropertyChangedCallback(MDILayoutPropertyChangedCallback)
            )
        );

        #endregion
        #region 활성 MDI 자식 속성 - ActiveMDIChildProperty

        /// <summary>
        /// 활성 MDI 자식 속성
        /// </summary>
        public static readonly DependencyProperty ActiveMDIChildProperty = DependencyProperty.Register
        (
            "ActiveMDIChild",
            typeof(MDIChild),
            typeof(MDIContainer),
            new UIPropertyMetadata
            (
                null,
                new PropertyChangedCallback(ActiveMDIChildPropertyChangedCallback)
            )
        );

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Internal

        #region 버튼 패널 속성 - ButtonPanelProperty

        /// <summary>
        /// 버튼 패널 속성
        /// </summary>
        internal static readonly DependencyProperty ButtonPanelProperty = DependencyProperty.Register
        (
            "ButtonPanel",
            typeof(Panel),
            typeof(MDIContainer),
            new UIPropertyMetadata
            (
                null,
                new PropertyChangedCallback(ButtonPanelPropertyChangedCallback)
            )
        );

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 현재 리소스 딕셔너리
        /// </summary>
        private static ResourceDictionary _currentResourceDictionary;

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 윈도우 오프셋
        /// </summary>
        private const int WindowOffset = 25;

        /// <summary>
        /// 윈도우 캔버스
        /// </summary>
        private Canvas windowCanvas;

        /// <summary>
        /// 메뉴 테두리
        /// </summary>
        private Border menuBorder;

        /// <summary>
        /// 버튼 테두리
        /// </summary>
        private Border buttonBorder;

        /// <summary>
        /// 상위 패널
        /// </summary>
        private Panel topPanel;

        /// <summary>
        /// 윈도우 오프셋
        /// </summary>
        private double windowOffset;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 테마 - Theme

        /// <summary>
        /// 테마
        /// </summary>
        public ThemeType Theme
        {
            get
            {
                return (ThemeType)GetValue(ThemeProperty);
            }
            set
            {
                SetValue(ThemeProperty, value);
            }
        }

        #endregion
        #region 메뉴 - Menu

        /// <summary>
        /// 메뉴
        /// </summary>
        public UIElement Menu
        {
            get
            {
                return (UIElement)GetValue(MenuProperty);
            }
            set
            {
                SetValue(MenuProperty, value);
            }
        }

        #endregion
        #region MDI 레이아웃 - MDILayout

        /// <summary>
        /// MDI 레이아웃
        /// </summary>
        public MDILayout MDILayout
        {
            get
            {
                return (MDILayout)GetValue(MDILayoutProperty);
            }
            set
            {
                SetValue(MDILayoutProperty, value);
            }
        }
        
        #endregion
        #region 활성 MDI 자식 - ActiveMDIChild

        /// <summary>
        /// 활성 MDI 자식
        /// </summary>
        public MDIChild ActiveMDIChild
        {
            get
            {
                return (MDIChild)GetValue(ActiveMDIChildProperty);
            }
            internal set
            {
                SetValue(ActiveMDIChildProperty, value);
            }
        }

        #endregion

        #region 자식 컬렉션 - Children

        /// <summary>
        /// 자식 컬렉션
        /// </summary>
        public ObservableCollection<MDIChild> Children { get; set; }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Internal

        #region 버튼 패널 - ButtonPanel

        /// <summary>
        /// 버튼 패널
        /// </summary>
        internal Panel ButtonPanel
        {
            get
            {
                return (Panel)GetValue(ButtonPanelProperty);
            }
            set
            {
                SetValue(ButtonPanelProperty, value);
            }
        }

        #endregion
        #region 내부 높이 - InnerHeight

        /// <summary>
        /// 내부 높이
        /// </summary>
        internal double InnerHeight
        {
            get
            {
                return ActualHeight - this.topPanel.ActualHeight;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - MDIContainer()

        /// <summary>
        /// 생성자
        /// </summary>
        public MDIContainer()
        {
            Background = Brushes.DarkGray;
            Focusable  = false;
            IsTabStop  = false;
            Children   = new ObservableCollection<MDIChild>();

            Children.CollectionChanged += Children_CollectionChanged;

            Grid grid = new Grid();

            grid.RowDefinitions.Add(new RowDefinition { Height = GridLength.Auto });
            grid.RowDefinitions.Add(new RowDefinition());

            this.topPanel = new DockPanel { Background = SystemColors.MenuBrush };

            this.topPanel.Children.Add(this.menuBorder = new Border());

            DockPanel.SetDock(this.menuBorder, Dock.Left);

            this.topPanel.Children.Add(this.buttonBorder = new Border());

            DockPanel.SetDock(this.buttonBorder, Dock.Right);

            this.topPanel.SizeChanged += UserControl_SizeChanged;

            this.topPanel.Children.Add(new UIElement());

            grid.Children.Add(this.topPanel);

            ScrollViewer scrollViewer = new ScrollViewer
            {
                Content                       = this.windowCanvas = new Canvas(),
                HorizontalScrollBarVisibility = ScrollBarVisibility.Auto,
                VerticalScrollBarVisibility   = ScrollBarVisibility.Auto
            };

            grid.Children.Add(scrollViewer);

            Grid.SetRow(scrollViewer, 1);

            Content = grid;

            if(Environment.OSVersion.Version.Major > 5)
            {
                ThemePropertyChangedCallback
                (
                    this,
                    new DependencyPropertyChangedEventArgs(ThemeProperty, Theme, ThemeType.Aero)
                );
            }
            else
            {
                ThemePropertyChangedCallback
                (
                    this,
                    new DependencyPropertyChangedEventArgs(ThemeProperty, Theme, ThemeType.Luna)
                );
            }

            Loaded      += UserControl_Loaded;
            SizeChanged += UserControl_SizeChanged;
            KeyDown     += UserControl_KeyDown;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private
        ////////////////////////////////////////////////////////////////////// Function

        #region 테마 속성 변경시 콜백 처리하기 - ThemePropertyChangedCallback(d, e)

        /// <summary>
        /// 테마 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void ThemePropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            MDIContainer container = d as MDIContainer;

            ThemeType themeType = (ThemeType)e.NewValue;

            bool maximumMode = container.ActiveMDIChild != null && container.ActiveMDIChild.WindowState == WindowState.Maximized;

            if(maximumMode)
            {
                container.ActiveMDIChild.WindowState = WindowState.Normal;
            }

            if(_currentResourceDictionary != null)
            {
                Application.Current.Resources.MergedDictionaries.Remove(_currentResourceDictionary);
            }

            switch(themeType)
            {
                case ThemeType.Luna :

                    Application.Current.Resources.MergedDictionaries.Add
                    (
                        _currentResourceDictionary = new ResourceDictionary
                        {
                            Source = new Uri
                            (
                                "/TestLibrary;component/Themes/Luna.xaml",
                                UriKind.Relative
                            )
                        }
                    );

                    break;

                case ThemeType.Aero :

                    Application.Current.Resources.MergedDictionaries.Add
                    (
                        _currentResourceDictionary = new ResourceDictionary
                        {
                            Source = new Uri
                            (
                                "/TestLibrary;component/Themes/Aero.xaml",
                                UriKind.Relative
                            )
                        }
                    );

                    break;
            }
        }

        #endregion
        #region 메뉴 속성 변경시 콜백 처리하기 - MenuPropertyChangedCallback(d, e)

        /// <summary>
        /// 메뉴 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void MenuPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            MDIContainer container = d as MDIContainer;

            UIElement menu = e.NewValue as UIElement;

            container.menuBorder.Child = menu;
        }

        #endregion
        #region MDI 레이아웃 속성 변경시 콜백 처리하기 - MDILayoutPropertyChangedCallback(d, e)

        /// <summary>
        /// MDI 레이아웃 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void MDILayoutPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            MDIContainer container = d as MDIContainer;

            MDILayout layout = (MDILayout)e.NewValue;

            if(layout == MDILayout.ArrangeIcons || container.Children.Count < 1)
            {
                return;
            }

            List<MDIChild> minimizedWindowList = new List<MDIChild>();
            List<MDIChild> normalWindowList    = new List<MDIChild>();

            foreach(MDIChild child in container.Children)
            {
                switch(child.WindowState)
                {
                    case WindowState.Minimized :

                        minimizedWindowList.Add(child);

                        break;

                    case WindowState.Maximized :

                        child.WindowState = WindowState.Normal;

                        normalWindowList.Add(child);

                        break;

                    default :

                        normalWindowList.Add(child);

                        break;
                }
            }

            minimizedWindowList.Sort(new MDIChildComparer());
            normalWindowList.Sort(new MDIChildComparer());

            double containerHeight = container.InnerHeight;

            for(int i = 0; i < minimizedWindowList.Count; i++)
            {
                MDIChild child = minimizedWindowList[i];

                int capacity = Convert.ToInt32(container.ActualWidth) / MDIChild.MinimizedWidth;
                int row      = i / capacity + 1;
                int column   = i % capacity;

                containerHeight = container.InnerHeight - MDIChild.MinimizedHeight * row;

                double newLeft = MDIChild.MinimizedWidth * column;

                child.Position = new Point(newLeft, containerHeight);
            }

            switch(layout)
            {
                case MDILayout.Cascade :
                {
                    double newWidth     = container.ActualWidth * 0.58;
                    double newHeight    = containerHeight * 0.67;
                    double windowOffset = 0;

                    foreach(MDIChild child in normalWindowList)
                    {
                        if(child.Resizable)
                        {
                            child.Width  = newWidth;
                            child.Height = newHeight;
                        }

                        child.Position = new Point(windowOffset, windowOffset);

                        windowOffset += WindowOffset;

                        if(windowOffset + child.Width > container.ActualWidth)
                        {
                            windowOffset = 0;
                        }

                        if(windowOffset + child.Height > containerHeight)
                        {
                            windowOffset = 0;
                        }
                    }

                    break;
                }
                case MDILayout.TileHorizontal :
                {
                    int column = (int)Math.Sqrt(normalWindowList.Count);
                    int row    = normalWindowList.Count / column;

                    List<int> columnList = new List<int>(); // windows per column

                    for(int i = 0; i < column; i++)
                    {
                        if(normalWindowList.Count % column > column - i - 1)
                        {
                            columnList.Add(row + 1);
                        }
                        else
                        {
                            columnList.Add(row);
                        }
                    }

                    double newWidth   = container.ActualWidth / column;
                    double newHeight  = containerHeight / columnList[0];
                    double offsetTop  = 0;
                    double offsetLeft = 0;

                    for(int i = 0, columnIndex = 0, previousCount = 0; i < normalWindowList.Count; i++)
                    {
                        if(i >= previousCount + columnList[columnIndex])
                        {
                            previousCount += columnList[columnIndex++];

                            offsetLeft += newWidth;

                            offsetTop = 0;

                            newHeight = containerHeight / columnList[columnIndex];
                        }

                        MDIChild child = normalWindowList[i];

                        if(child.Resizable)
                        {
                            child.Width  = newWidth;
                            child.Height = newHeight;
                        }

                        child.Position = new Point(offsetLeft, offsetTop);

                        offsetTop += newHeight;
                    }

                    break;
                }
                case MDILayout.TileVertical :
                {
                    int row    = (int)Math.Sqrt(normalWindowList.Count);
                    int column = normalWindowList.Count / row;

                    List<int> columnList = new List<int>();

                    for(int i = 0; i < column; i++)
                    {
                        if(normalWindowList.Count % column > column - i - 1)
                        {
                            columnList.Add(row + 1);
                        }
                        else
                        {
                            columnList.Add(row);
                        }
                    }

                    double newWidth   = container.ActualWidth / column;
                    double newHeight  = containerHeight / columnList[0];
                    double offsetTop  = 0;
                    double offsetLeft = 0;

                    for(int i = 0, columnIndex = 0, previousCount = 0; i < normalWindowList.Count; i++)
                    {
                        if(i >= previousCount + columnList[columnIndex])
                        {
                            previousCount += columnList[columnIndex++];

                            offsetLeft += newWidth;

                            offsetTop = 0;

                            newHeight = containerHeight / columnList[columnIndex];
                        }

                        MDIChild child = normalWindowList[i];

                        if(child.Resizable)
                        {
                            child.Width  = newWidth;
                            child.Height = newHeight;
                        }

                        child.Position = new Point(offsetLeft, offsetTop);

                        offsetTop += newHeight;
                    }

                    break;
                }
            }

            container.InvalidateSize();

            container.MDILayout = MDILayout.ArrangeIcons;
        }

        #endregion
        #region 활성 MDI 자식 속성 변경시 콜백 처리하기 - ActiveMDIChildPropertyChangedCallback(d, e)

        /// <summary>
        /// 활성 MDI 자식 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void ActiveMDIChildPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            MDIContainer container = d as MDIContainer;

            MDIChild newChild = e.NewValue as MDIChild;
            MDIChild oldChild = e.OldValue as MDIChild;

            if(newChild == null || newChild == oldChild)
            {
                return;
            }

            if(oldChild != null &&  oldChild.WindowState == WindowState.Maximized)
            {
                newChild.WindowState = WindowState.Maximized;
            }

            int maximuZIndex = 0;

            for(int i = 0; i < container.Children.Count; i++)
            {
                int zIndex = Panel.GetZIndex(container.Children[i]);

                if(zIndex > maximuZIndex)
                {
                    maximuZIndex = zIndex;
                }

                if(container.Children[i] != newChild)
                {
                    container.Children[i].Focused = false;
                }
                else
                {
                    newChild.Focused = true;
                }
            }

            Panel.SetZIndex(newChild, maximuZIndex + 1);

            if(container.MDIChildTitleChanged != null)
            {
                container.MDIChildTitleChanged(container, new RoutedEventArgs());
            }
        }

        #endregion
        #region 버튼 패널 속성 변경시 콜백 처리하기 - ButtonPanelPropertyChangedCallback(d, e)

        /// <summary>
        /// 버튼 패널 속성 변경시 콜백 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private static void ButtonPanelPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            MDIContainer container = d as MDIContainer;

            Panel panel = e.NewValue as Panel;

            container.buttonBorder.Child = panel;

            if(container.MDIChildTitleChanged != null)
            {
                container.MDIChildTitleChanged(container, new RoutedEventArgs());
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Internal
        ////////////////////////////////////////////////////////////////////// Function

        #region 크기 무효화하기 - InvalidateSize()

        /// <summary>
        /// 크기 무효화하기
        /// </summary>
        internal void InvalidateSize()
        {
            Point largestPoint = new Point(0, 0);

            for(int i = 0; i < Children.Count; i++)
            {
                MDIChild child = Children[i];

                Point farPosition = new Point
                (
                    child.Position.X + child.Width,
                    child.Position.Y + child.Height
                );

                if(farPosition.X > largestPoint.X)
                {
                    largestPoint.X = farPosition.X;
                }

                if(farPosition.Y > largestPoint.Y)
                {
                    largestPoint.Y = farPosition.Y;
                }
            }

            if(this.windowCanvas.Width != largestPoint.X)
            {
                this.windowCanvas.Width = largestPoint.X;
            }

            if(this.windowCanvas.Height != largestPoint.Y)
            {
                this.windowCanvas.Height = largestPoint.Y;
            }
        }

        #endregion
        #region 상위 자식 구하기 - GetTopChild()

        /// <summary>
        /// 상위 자식 구하기
        /// </summary>
        /// <returns>상위 자식</returns>
        internal MDIChild GetTopChild()
        {
            if(Children.Count < 1)
            {
                return null;
            }

            int index         = 0;
            int maximumZIndex = Panel.GetZIndex(Children[0]);

            for(int i = 1, zIndex; i < Children.Count; i++)
            {
                zIndex = Panel.GetZIndex(Children[i]);

                if(zIndex > maximumZIndex)
                {
                    maximumZIndex = zIndex;
                    index         = i;
                }
            }

            return Children[index];
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Private
        ////////////////////////////////////////////////////////////////////// Event

        #region 사용자 컨트롤 로드시 처리하기 - UserControl_Loaded(sender, e)

        /// <summary>
        /// 사용자 컨트롤 로드시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void UserControl_Loaded(object sender, RoutedEventArgs e)
        {
            Window window = Window.GetWindow(this);

            if(window != null)
            {
                window.Activated   += window_Activated;
                window.Deactivated += window_Deactivated;
            }

            this.windowCanvas.Width  = this.windowCanvas.ActualWidth;
            this.windowCanvas.Height = this.windowCanvas.ActualHeight;

            this.windowCanvas.VerticalAlignment   = VerticalAlignment.Top;
            this.windowCanvas.HorizontalAlignment = HorizontalAlignment.Left;

            InvalidateSize();
        }

        #endregion
        #region 사용자 컨트롤 크기 변경시 처리하기 - UserControl_SizeChanged(sender, e)

        /// <summary>
        /// 사용자 컨트롤 크기 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void UserControl_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            if(Children.Count == 0)
            {
                return;
            }

            for(int i = 0; i < Children.Count; i++)
            {
                MDIChild child = Children[i];

                if(child.WindowState == WindowState.Maximized)
                {
                    child.Width  = ActualWidth;
                    child.Height = ActualHeight;
                }

                if(child.WindowState == WindowState.Minimized)
                {
                    child.Position = new Point
                    (
                        child.Position.X,
                        child.Position.Y + e.NewSize.Height - e.PreviousSize.Height
                    );
                }
            }
        }

        #endregion
        #region 사용자 컨트롤 키 DOWN 처리하기 - UserControl_KeyDown(sender, e)

        /// <summary>
        /// 사용자 컨트롤 키 DOWN 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void UserControl_KeyDown(object sender, KeyEventArgs e)
        {
            MDIContainer container = (MDIContainer)sender;

            if(container.Children.Count < 2)
            {
                return;
            }

            switch(e.Key)
            {
                case Key.Tab :

                    if(e.KeyboardDevice.Modifiers == ModifierKeys.Control)
                    {
                        int minimumZIndex = Panel.GetZIndex(container.Children[0]);

                        foreach(MDIChild child in container.Children)
                        {
                            if(Panel.GetZIndex(child) < minimumZIndex)
                            {
                                minimumZIndex = Panel.GetZIndex(child);
                            }
                        }

                        Panel.SetZIndex(container.GetTopChild(), minimumZIndex - 1);

                        container.GetTopChild().Focus();

                        e.Handled = true;
                    }

                    break;
            }
        }

        #endregion

        #region 자식 컬렉션 변경시 처리하기 - Children_CollectionChanged(sender, e)

        /// <summary>
        /// 자식 컬렉션 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void Children_CollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
        {
            switch(e.Action)
            {
                case NotifyCollectionChangedAction.Add :
                {
                    MDIChild child    = Children[e.NewStartingIndex];
                    MDIChild topChild = ActiveMDIChild;

                    if(topChild != null && topChild.WindowState == WindowState.Maximized)
                    {
                        child.Loaded += (sender2, e2) => child.WindowState = WindowState.Maximized;
                    }

                    child.Loaded += (sender2, e2) => ActiveMDIChild = child;

                    if(child.Position.X < 0 || child.Position.Y < 0)
                    {
                        child.Position = new Point(this.windowOffset, this.windowOffset);
                    }

                    this.windowCanvas.Children.Add(child);

                    this.windowOffset += WindowOffset;

                    if(this.windowOffset + child.Width > ActualWidth)
                    {
                        this.windowOffset = 0;
                    }

                    if(this.windowOffset + child.Height > ActualHeight)
                    {
                        this.windowOffset = 0;
                    }

                    break;
                }
                case NotifyCollectionChangedAction.Remove :
                {
                    MDIChild oldChild = (MDIChild)e.OldItems[0];

                    this.windowCanvas.Children.Remove(oldChild);

                    MDIChild newChild = GetTopChild();

                    ActiveMDIChild = newChild;

                    if(newChild != null && oldChild.WindowState == WindowState.Maximized)
                    {
                        newChild.WindowState = WindowState.Maximized;
                    }

                    break;
                }
                case NotifyCollectionChangedAction.Reset :
                {
                    this.windowCanvas.Children.Clear();

                    break;
                }
            }

            InvalidateSize();
        }

        #endregion

        #region 윈도우 활성시 처리하기 - window_Activated(object sender, EventArgs e)

        /// <summary>
        /// 윈도우 활성시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void window_Activated(object sender, EventArgs e)
        {
            if(ActiveMDIChild == null)
            {
                return;
            }

            ActiveMDIChild.Focused = true;
        }

        #endregion
        #region 윈도우 비활성시 처리하기 - window_Deactivated(sender, e)

        /// <summary>
        /// 윈도우 비활성시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void window_Deactivated(object sender, EventArgs e)
        {
            if(Children.Count == 0)
            {
                return;
            }

            for(int i = 0; i < Children.Count; i++)
            {
                if(Children[i].WindowState != WindowState.Maximized)
                {
                    Children[i].Focused = false;
                }
            }
        }

        #endregion
    }
}