﻿using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media.Imaging;

using TestLibrary;

namespace TestProject
{
    /// <summary>
    /// 메인 윈도우
    /// </summary>
    public partial class MainWindow : Window
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 원본 제목
        /// </summary>
        private string originalTitle;

        /// <summary>
        /// 순번
        /// </summary>
        private int sequence = 1;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - MainWindow()

        /// <summary>
        /// 생성자
        /// </summary>
        public MainWindow()
        {
            InitializeComponent();

            this.originalTitle = Title;

            this.container.Children.CollectionChanged += (sender, e) => BuildWindowMenuItem();

            this.container.MDIChildTitleChanged += container_MDIChildTitleChanged;

            this.container.Children.Add
            (
                new MDIChild
                {
                    Title    = "코드를 사용한 빈 윈도우",
                    Icon     = new BitmapImage(new Uri("/IMAGE/OriginalLogo.png", UriKind.Relative)),
                    Position = new Point(160, 160)
                }
            );

            this.container.Children.Add
            (
                new MDIChild
                {
                    Title    = "코드를 사용한 윈도우",
                    Content  = new SampleControl(),
                    Position = new Point(190, 190)
                }
            );
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Private
        //////////////////////////////////////////////////////////////////////////////// Event

        #region 일반 윈도우 추가 메뉴 항목 클릭시 처리하기 - addWindowMenuItem_Click(sender, e)

        /// <summary>
        /// 일반 윈도우 추가 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void addWindowMenuItem_Click(object sender, RoutedEventArgs e)
        {
            this.container.Children.Add
            (
                new MDIChild
                {
                    Content = new Label{ Content = "일반 윈도우" },
                    Title   = $"윈도우 {this.sequence++}"
                }
            );
        }

        #endregion
        #region 고정 윈도우 추가 메뉴 항목 클릭시 처리하기 - addFixedWindowMenuItem_Click(sender, e)

        /// <summary>
        /// 고정 윈도우 추가 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void addFixedWindowMenuItem_Click(object sender, RoutedEventArgs e)
        {
            this.container.Children.Add
            (
                new MDIChild
                {
                    Content   = new Label { Content = "고정 너비 윈도우" },
                    Title     = $"윈도우 {this.sequence++}",
                    Resizable = false
                }
            );
        }

        #endregion
        #region 스크롤 윈도우 추가 메뉴 항목 클릭시 처리하기 - addScrollWindowMenuItem_Click(sender, e)

        /// <summary>
        /// 스크롤 윈도우 추가 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void addScrollWindowMenuItem_Click(object sender, RoutedEventArgs e)
        {
            StackPanel stackPanel = new StackPanel
            {
                Orientation = Orientation.Vertical
            };

            stackPanel.Children.Add
            (
                new TextBlock
                {
                    Margin = new Thickness(5),
                    Text   = "스크롤 윈도우"
                }
            );

            stackPanel.Children.Add
            (
                new ComboBox
                {
                    Margin = new Thickness(20),
                    Width  = 300
                }
            );

            ScrollViewer scrollViewer = new ScrollViewer
            {
                HorizontalScrollBarVisibility = ScrollBarVisibility.Auto,
                Content                       = stackPanel
            };

            this.container.Children.Add
            (
                new MDIChild
                {
                    Content = scrollViewer,
                    Title = $"윈도우 {this.sequence++}"
                }
            );
        }

        #endregion

        #region 일반 메뉴 항목 클릭시 처리하기 - genericMenuItem_Click(sender, e)

        /// <summary>
        /// 일반 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void genericMenuItem_Click(object sender, RoutedEventArgs e)
        {
            this.genericMenuItem.IsChecked = true;
            this.lunaMenuItem.IsChecked    = false;
            this.aeroMenuItem.IsChecked    = false;

            this.container.Theme = ThemeType.Generic;
        }

        #endregion
        #region 루나 메뉴 항목 클릭시 처리하기 - lunaMenuItem_Click(sender, e)

        /// <summary>
        /// 루나 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void lunaMenuItem_Click(object sender, RoutedEventArgs e)
        {
            this.genericMenuItem.IsChecked = false;
            this.lunaMenuItem.IsChecked    = true;
            this.aeroMenuItem.IsChecked    = false;

            this.container.Theme = ThemeType.Luna;
        }

        #endregion
        #region 에어로 메뉴 항목 클릭시 처리하기 - aeroMenuItem_Click(sender, e)

        /// <summary>
        /// 에어로 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void aeroMenuItem_Click(object sender, RoutedEventArgs e)
        {
            this.genericMenuItem.IsChecked = false;
            this.lunaMenuItem.IsChecked    = false;
            this.aeroMenuItem.IsChecked    = true;

            this.container.Theme = ThemeType.Aero;
        }

        #endregion

        #region 컨테이너 MDI 자식 제목 변경시 처리하기 - container_MDIChildTitleChanged(sender, e)

        /// <summary>
        /// 컨테이너 MDI 자식 제목 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void container_MDIChildTitleChanged(object sender, RoutedEventArgs e)
        {
            if(this.container.ActiveMDIChild != null && this.container.ActiveMDIChild.WindowState == WindowState.Maximized)
            {
                Title = this.originalTitle + " - [" + this.container.ActiveMDIChild.Title + "]";
            }
            else
            {
                Title = this.originalTitle;
            }
        }

        #endregion

        #region 최소화 상자 비활성화 버튼 클릭시 처리하기 - disableMinimizeButton_Click(sender, e)

        /// <summary>
        /// 최소화 상자 비활성화 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void disableMinimizeButton_Click(object sender, RoutedEventArgs e)
        {
            this.window1.MinimizeBox = false;
        }

        #endregion
        #region 최소화 상자 활성화 버튼 클릭시 처리하기 - enableMinimizeButton_Click(sender, e)

        /// <summary>
        /// 최소화 상자 활성화 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void enableMinimizeButton_Click(object sender, RoutedEventArgs e)
        {
            this.window1.MinimizeBox = true;
        }

        #endregion
        #region 최대화 상자 비활성화 버튼 클릭시 처리하기 - disableMaximizeButton_Click(sender, e)

        /// <summary>
        /// 최대화 상자 비활성화 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void disableMaximizeButton_Click(object sender, RoutedEventArgs e)
        {
            this.window1.MaximizeBox = false;
        }

        #endregion
        #region 최대화 상자 활성화 버튼 클릭시 처리하기 - enableMaximizeButton_Click(sender, e)

        /// <summary>
        /// 최대화 상자 활성화 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void enableMaximizeButton_Click(object sender, RoutedEventArgs e)
        {
            this.window1.MaximizeBox = true;
        }

        #endregion
        #region 닫기 상자 비활성화 버튼 클릭시 처리하기 - disableCloseButton_Click(sender, e)

        /// <summary>
        /// 닫기 상자 비활성화 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void disableCloseButton_Click(object sender, RoutedEventArgs e)
        {
            this.window1.CloseBox = false;
        }

        #endregion
        #region 닫기 상자 활성화 버튼 클릭시 처리하기 - enableCloseButton_Click(sender, e)

        /// <summary>
        /// 닫기 상자 활성화 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void enableCloseButton_Click(object sender, RoutedEventArgs e)
        {
            this.window1.CloseBox = true;
        }

        #endregion
        #region 아이콘 표시 버튼 클릭시 처리하기 - showIconButton_Click(sender, e)

        /// <summary>
        /// 아이콘 표시 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void showIconButton_Click(object sender, RoutedEventArgs e)
        {
            this.window1.ShowIcon = true;
        }

        #endregion
        #region 아이콘 숨김 버튼 클릭시 처리하기 - hideIconButton_Click(sender, e)

        /// <summary>
        /// 아이콘 숨김 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void hideIconButton_Click(object sender, RoutedEventArgs e)
        {
            this.window1.ShowIcon = false;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Function

        #region 윈도우 메뉴 항목 만들기 - BuildWindowMenuItem()

        /// <summary>
        /// 윈도우 메뉴 항목 만들기
        /// </summary>
        private void BuildWindowMenuItem()
        {
            this.windowMenuItem.Items.Clear();

            MenuItem menuItem;

            for(int i = 0; i < this.container.Children.Count; i++)
            {
                MDIChild child = this.container.Children[i];

                menuItem = new MenuItem { Header = child.Title };

                menuItem.Click += (s, e) => child.Focus();

                this.windowMenuItem.Items.Add(menuItem);
            }

            this.windowMenuItem.Items.Add(new Separator());
            this.windowMenuItem.Items.Add(menuItem = new MenuItem { Header = "계단식" });

            menuItem.Click += (s, e) => container.MDILayout = MDILayout.Cascade;

            this.windowMenuItem.Items.Add(menuItem = new MenuItem { Header = "수평" });

            menuItem.Click += (s, e) => container.MDILayout = MDILayout.TileHorizontal;

            this.windowMenuItem.Items.Add(menuItem = new MenuItem { Header = "수직" });

            menuItem.Click += (s, e) => container.MDILayout = MDILayout.TileVertical;

            this.windowMenuItem.Items.Add(new Separator());

            this.windowMenuItem.Items.Add(menuItem = new MenuItem { Header = "모두 닫기" });

            menuItem.Click += (s, e) => container.Children.Clear();
        }

        #endregion
    }
}